#!/bin/bash
#* Licensed Materials - Property of IBM
#*
#* OCO Source Materials
#*
#* (C) Copyright IBM Corp. 2011, 2012 All Rights Reserved
#*
# hostvolattach.sh [-l] [-m] [--ids <WWID>:<LUN>~<address> ...]
#   version: 2.4.3 [1/04/2013]
#   author:  csky
#
# !Script only supported on RHEL 6!
#
#    Description: Scan scsi bus HBA's for fresh set of volumes. Optionally look
#                 for a specific set. Dump out info for the set or all disks found.
#                 Optionally check first for MPIO stable paths (e.g. /dev/mapper/...)
#                 Call a fixing script if any of the specific set of volumes are
#                 not located.
#    Options:
#        -l    long (verbose)
#        -m    multi-path enabled
#       --ids   All arguments following this option are treated as a specific volume
#               reference, consisting of a WWID (DeviceID) and LUN combination
#               separated by a ':'. If specified, then only those devices are fully
#               processed. Missing stable paths can be fixed if the -f option is
#               specified in this case. If the LUN is unknown, pass -1. <address>
#               is the external named path to the storage (not the host view).
# 

cmdwait=15
scanwait=120
checkwait=500
vmcdir="/var/opt/ibm/vmc"
fixlogfile="${vmcdir}/outsancheck.trace"
alogfile="${vmcdir}/outattach.trace"
argids=N
long=false
fixmode="id-enabled"
let innertries=1
err_codes=""

#subroutines
log_msg(){
	local msg=$1
	echo "[$$] $msg" >> $alogfile
	if [ "$long" = "true" ]
	then
		echo "$msg"
	fi
}

log_err(){
	local msg=$1
	dstamp=$(date)
	echo "[$$]$dstamp: $msg" >> $alogfile
	# 1>&2 sends stdout to stderr
	echo "[$$]$dstamp: $msg" 1>&2
}

log_print(){
    local msg=$1
	echo "[$$] $msg" >> $alogfile
	echo "[$$] $msg"
}

get_scsiid(){
	log_msg "  scsi_id --page=0x83 --whitelisted --device=$1 | cut -d ' ' -f 1"
	scsiid=$(timeout -s SIGKILL $cmdwait scsi_id --page=0x83 --whitelisted --device=$1 | cut -d ' ' -f 1)
	local sret=$?
	if [ "$sret" = "0" ]
	then
		scsiid=${scsiid##[ ]*}
		return 0
	else
		scsiid="-"
		err_codes="$err_codes s${sret}\n"
		log_err "[scsi_id return for device=$1  **Command failed. [rc=$sret]"
		return $sret
	fi
}

dd_read(){
	log_msg "  dd if=$1 of=/dev/null bs=32 count=1 2>&1"
	chkread=$(timeout -s SIGKILL $cmdwait dd if=$1 of=/dev/null bs=32 count=1 2>&1)
	local sret=$?
	if [ "$sret" != "0" ]
	then
		err_codes="$err_codes d${sret}\n"
	fi
	log_msg "  dd reading $1 [rc=$sret]   output=$chkread"
	return $sret
}

# The "locking" done in this function is not to provide true synchronization.
# It just helps alleviate the load if many attach requests come in together so
# that the looping and processing through the volumes does not happen at once and
# bog down the system.
sancheck(){
	local sargs=$1
	local logto=$2
	local dstamp=$(date)
	local loop=0
	log_msg "Check for lck file. $sargs [$3]"
	if [ -f $vmcdir/lck.sancheck ]
	then
		log_msg "$dstamp ***sancheck lock file FOUND. Wait till it is gone or too old or starved out..."
		while [ -f $vmcdir/lck.sancheck ]
		do
			old=$(( (`date +%s` - `stat -L --format %Y $vmcdir/lck.sancheck`) > (30*60) ))
			log_msg "lck check: old=$old"
			if [ "$old" = "1" ]
			then
				log_msg "Removing old lock file and continuing..."
				rm -f $vmcdir/lck.sancheck
			fi
			contents=$(cat $vmcdir/lck.sancheck)
			sleep 1
			let loop=loop+1
			if [ $loop -gt $checkwait ]
			then
				log_msg "!Starved out waiting for sancheck lock! Loop=$loop"
				break
			fi
		done
		if [[ "$sargs" == *ids* ]]
		then
			log_msg "Specific vols passed to fix. Continue with sancheck..."
		else
			if [[ "$contents" == *ids* ]]
			then
				log_msg "Continue with general volume cleanup requested. Prior cleanup was for specific vols."
			else
				log_msg "Prior cleanup requested was general and since this is general, just return."
				return 0
			fi
		fi
		marker=n
		[ -f "$vmcdir/stale.sancheck" ] && marker=y
		if [ "$3" = "stale" -a "$marker" = "n" ]
		then
			log_msg "Stale marker no longer exists. Just return."
			return 0
		else
			log_msg "Stale marker still there. Continue with cleanup."
		fi
	fi

	echo $sargs >> $vmcdir/lck.sancheck
	local lines=$(cat $vmcdir/lck.sancheck | wc -l)
	log_msg "Touched lock file: $vmcdir/lck.sancheck  [lines=$lines]"
	dstamp=$(date)
	log_msg "  $dstamp: timeout -s SIGKILL 150 $vmcdir/sancheck.sh $sargs"
	timeout -s SIGKILL $checkwait $vmcdir/sancheck.sh $sargs >> $logto
	local sret=$?
	dstamp=$(date)
	echo "  $dstamp   *** sancheck numfixed/rc=$sret ***" >> $logto
	log_msg "  $dstamp   *** sancheck numfixed/rc=$sret ***"
	rm -f $vmcdir/lck.sancheck
	log_msg "Removed lock file: $vmcdir/lck.sancheck"
	if [ "$3" = "stale" ]
	then
		rm -f "$vmcdir/stale.sancheck"
		log_msg "Removed any $vmcdir/stale.sancheck"
	fi
	return $sret
}

hba_scan(){
	local dstamp=$(date)
	# $1 is ignored
	local hbas="$2"
	# Do we want to synchronize the scans and/or prevent scans that happened within the last couple seconds?
	log_msg "  $dstamp: timeout -s SIGKILL $scanwait echo '- - -' | tee $hbas > /dev/null"
	timeout -s SIGKILL $scanwait echo '- - -' | tee $hbas > /dev/null
	local sret=$?
	dstamp=$(date)
	log_msg "  $dstamp *** hba scsi bus scan   rc=$sret ***"
	if [ "$sret" != 0 ]
	then
		err_codes="$err_codes d${sret}\n"
    fi
	return $sret
}

get_vol_inventory(){
	dstamp=$(date)
	log_msg "General volume inventory collection. First identify known volumes. $dstamp"
	idlistfile=$(mktemp $vmcdir/idlist.XXXXXXXX)
	for fcdir in $(ls -1 /sys/class/fc_transport)
	do
		target=$(echo $fcdir | cut -d : -f 1)
		channel=$(echo $fcdir | cut -d : -f 2)
		controller=$(echo $fcdir | cut -d : -f 3)
		hostnum=${target#target}
		#echo "Dir: /sys/class/fc_transport/$fcdir/device"
		for file in $(ls -1 /sys/class/fc_transport/$fcdir/device | egrep "^$hostnum:$channel:$controller:")
		do
			dlink=/sys/class/fc_transport/$fcdir/device/$file
			if [ -d $dlink/block ]
			then
				blockdev=$(ls $dlink/block)
				#echo "$file: $blockdev"
				devblockpath=/dev/$blockdev
				get_scsiid $devblockpath
				ret=$?
				if [ ${#scsiid} -gt 2 -o "$ret" != "0" ]
				then
					volid=${scsiid:1}
					volid=${volid##[ ]*}
					if [ -n "$volid" ]
					then
						# Avoid duplicates so just take the first device with the wwid
						ogrep=$(grep $volid $idlistfile)
						rc=$?
						if [ "$rc" != "0" ]
						then
							mapline=$(awk -v ss="$volid" '$0 ~ ss {print $17,$13,$5}' $vmcdir/pvmap | sort -r | head -n 1)
							pvlun=$(echo $mapline | cut -d ' ' -f 3)
							if [ -n "$pvlun" ]
							then
								pvdevname=$(echo $mapline | cut -d ' ' -f 2)
								if [[ "$pvdevname" != /dev/* ]]; then
									# does not start with /dev/ so prior symbolic name is preferred
									devblockpath=$pvdevname
								fi
							else
								pvlun=-1
							fi
							echo "${volid}:$pvlun~$devblockpath" >> $idlistfile
						fi
					fi
				else
					log_msg "Skipping device $devblockpath because unsupported scsiid format or scsi_id err: id=$scsiid ret=$ret"
				fi
			fi
		done
	done
	volumes=$(cat $idlistfile)
	dstamp=$(date)
	log_msg "Volume inventory set: *** $dstamp ***"
	log_msg "$volumes"
	rm -f $idlistfile
}

found_stable(){
	local pathpre=$1
	local cutfield=$2
	log_msg "Stable device: $stable"
	dev=$(echo $stable | cut -d '/' -f $cutfield)
	devpath=/dev/$dev
	stablename=$(echo $stable | cut -d ' ' -f 1)
	stablepath=$pathpre/$stablename
	log_msg "Stable path: $stablepath"
	volfound=Y
}

find_specified_vols(){
	let numvols=0
	let numfound=0
	listnotfound=""
	dstamp=$(date)
	log_msg "find_specified_vols: $dstamp"
	if [ "$pvmapfile" != "" ]
	then
		rm -f $pvmapfile
	fi
	pvmapfile=$(mktemp $vmcdir/pvmap.XXXXXXXX)
	# Physical Volume loop - looking for physical volumes
	for vol in $volumes
	do
		log_msg "======================================================"
		log_msg "Volume: $vol"
		let numvols+=1
		volfound=N
		mpath=N
		health=OK
		# Split volume arguments in pieces
		argwwid=${vol%%:*}
		argvoli=${vol##*:}
		arglun=${argvoli%%~*}
		argname=${argvoli##*~}
		if [ "$arglun" = "$argwwid" ]
		then
			arglun="-1"
			argname="[unknown-address]"
		fi
		if [ "$argname" = "$arglun" ]
		then
			argname="[unknown-address]"
		fi
		# Search in priority order
		if [ "$multipath" = "Y" ]
		then
			# 1. for multipath, search /dev/mapper
			stable=$(ls -l /dev/mapper | awk -v ss="[^p]$argwwid[^p]" '$0 ~ ss {print $9, $11}')
			if [ -n "$stable" ]
			then
				found_stable "/dev/mapper" "2"
				mpath=Y
			else
				# fallback to by-id, though this should be pretty unlikely
				stable=$(ls -l /dev/disk/by-id | awk -v ss="dm-uuid-mpath-[0-9]$argwwid" '$0 ~ ss {print $9, $11}')
				if [ -n "$stable" ]
				then
					found_stable "/dev/disk/by-id" "3"
					mpath=Y
				fi
			fi
		fi
		if [ "$volfound" = "N" ]
		then
			# wwn- prefix check (non-multipath)
			stable=$(ls -l /dev/disk/by-id | awk -v ss="wwn-0x$argwwid[^-]" '$0 ~ ss {print $9, $11}')
			if [ -n "$stable" ]
			then
				found_stable "/dev/disk/by-id" "3"
			else
				# fallback - allow any scsi type prefix that is a single digit, not just '3' ('scsi-.' below)
				stable=$(ls -l /dev/disk/by-id | awk -v ss="scsi-.$argwwid[^-]" '$0 ~ ss {print $9, $11}')
				if [ -n "$stable" ]
				then
					found_stable "/dev/disk/by-id" "3"
				fi
			fi
		fi
		if [ "$volfound" = "Y" ]
		then
			# Test device
			dd_read $devpath
			rc=$?
			if [ "$rc" != "0" ]
			then
				health="read-failure"
				# If the fix mode is "none", then no more volume 'fixing' will be attempted -
				# leave the volume as 'found' even though it cannot be read.
				if [ "$fixmode" != "none" ]
				then
					volfound=N
				fi
			fi

			let numfound+=1
			# Common 'disk' pieces
			blockdir=/sys/block/$dev
			blocks=$(cat $blockdir/size)
			if [ "$blocks" = "" ]
			then
				blocks="0"
			fi
			blocksize=$(cat $blockdir/queue/physical_block_size)
			if [ "$blocksize" = "" ]
			then
				blocksize="0"
			fi
			sizeGiB=$(echo "scale=3; ($blocks * $blocksize) / ( 1024 * 1024 * 1024 )" | bc)
			ronly=$(cat $blockdir/ro)
			access=$ronly
			if [ "$ronly" = "0" ]
			then
				#read/write
				access="3"
			else
				if [ "$ronly" != "1" ]
				then
					#unknown
					access="0"
				fi
			fi
			if [ "$mpath" = "Y" ]
			then
				# For multipath output, assume the "active" path is 3 lines below the wwid entry.
				scsidev=$(multipath -ll | grep -A 3 $argwwid | tail -n 1 | awk '/running/ {print $3}')
				# The scsidev (e.g. 1:0:2:3) is used to index into the sysfs device path
				dlink=/sys/class/scsi_device/$scsidev/device
			else
				# Non-multipath case. The Device dir is right underneath the $blockdir
				dlink=$blockdir/device
				scsidev=$(ls $dlink/scsi_device)
			fi
			
			# Common 'device' pieces
			scsi3dev=${scsidev%:*}
			lun=$(echo $scsidev | cut -d : -f 4)
			vendor=$(cat $dlink/vendor)
			vendor=${vendor%%[ ]*}
			model=$(cat $dlink/model)
			model=${model%%[ ]*}
			transportdir=/sys/class/fc_transport/target$scsi3dev
			if [ -d $transportdir ]
			then
				hostwwpn=$(cat $transportdir/port_name)
			else
				# probably should exclude these from inventory
				hostwwpn="NON-FC-TRANSPORT"
			fi
			
			# send upper case wwid back
			deviceID=$(echo $argwwid | tr '[a-z]' '[A-Z]')
			# validation
			if [[ "$argname" == /dev/* ]]; then
				argname=$devpath
			fi
			if [ "$hostwwpn" = "" ]
			then
				hostwwpn="unknown"
			fi
			if [ "$vendor" = "" ]
			then
				vendor="unknown"
			fi
			if [ "$model" = "" ]
			then
				model="unknown"
			fi
			if [ -n $devpath -a -b $devpath ]
			then
				curdate=$(stat -c %y $devpath | awk '{print $1 "_" $2}')
				curdate=${curdate%%.*}
			else
				curdate="unknown"
			fi
			vsuse="---"
			if [ "$numvses" != "0" ]
			then
				filematch=$(grep $stablepath /etc/libvirt/qemu/*.xml | head -n 1 | cut -d : -f 1)
				if [ -n "$filematch" ]
				then
					vsuse=${filematch%%.xml}
					vsuse=${vsuse##*/}
				fi
			fi
			# wwid  stable-path  devpath  scsi-addr  lun-in  host-lun  blocks  mpath  access  host-wwpn  vendor  model  external-address  block-size  sizeGiB  health  device-date  using-vs
			volout=$(printf '%-33s %-70s %12s %11s %4s %4s %10s %1s %1s %18s %8s %8s %s %s  %s  %s  %s  %s' $deviceID $stablepath $devpath $scsidev $arglun $lun $blocks $mpath "$access" $hostwwpn $vendor $model $argname $blocksize $sizeGiB $health $curdate $vsuse)
			if [ "$hostwwpn" != "NON-FC-TRANSPORT" ]
			then
				# Only add this one if FC transport device
				echo "$volout" >> $pvmapfile
			fi
			log_msg "  ------------- Found volume $numfound -------------------"
			log_msg "$volout"
		else
			log_msg "  **Volume not found: $argwwid  lun=$arglun  address=$argname"
			listnotfound="$listnotfound ${argwwid}:${arglun}~${argname}"
			# keep going to get full list to fix
		fi
		# On to next volume to find
	done
}
# end volume search method
# Process script arguments
if [ "$1" = "-l" -o "$2" = "-l" ]
then
	long=true
fi
multipath=N
if [ "$1" = "-m" -o "$2" = "-m" ]
then
	multipath=Y
fi

if [ "$1" = "--ids" ]
then
	shift 1
	argids=Y
else
	if [ "$2" = "--ids" ]
	then
		shift 2
		argids=Y
	else
		if [ "$3" = "--ids" ]
		then
			shift 3
			argids=Y
		fi
	fi
fi

# This folder should exist since normally the scripts are contained in it.
if [ -d $vmcdir ]
then
	log_msg ""
else
	mkdir -p $vmcdir
	rc=$?
	if [ "$long" = "true" ]
	then
		echo "Created directory $vmcdir. rc=$rc"
	fi
fi

if [ -f $alogfile ]
then
	ccrc=$(find $alogfile -size +3000)
	if [ "$ccrc" = "$alogfile" ]
	then
		echo "Cleanup old log file (wrap to backup)."
		mv -f $alogfile "${alogfile}.1"
	fi
fi
dstamp=$(date)
log_msg "$dstamp"
log_msg "long=$long   argids=$argids   multipath=$multipath"
touch $vmcdir/pvmap

if [ "$argids" = "Y" ]
then
	volumes="$@"
	log_msg "Volumes= $volumes"
fi

# Check HBA ports
numHosts=$(ls -1 /sys/class/fc_host/ | wc -l)
if [ "$numHosts" = "0" ]
then
	log_err "This endpoint does not appear to be attached to a Fibre Channel network. No entries in /sys/class/fc_host. Exiting."
	exit 1
fi

# compute FC HBAs
hbaHosts=$(ls /sys/class/fc_host/ | sed 's/host[0-9]*/\/sys\/class\/scsi_host\/&\/scan/g' | tr "\n" " ")

# Do the "light" scan of scsi interconnect (the only semi-safe way)
hba_scan "-" "$hbaHosts"

numOnline=$(cat /sys/class/fc_host/host*/port_state | grep -c Online)
if [ "$numOnline" = "0" ]
then
	log_err "None of the HBA ports have a status of Online."
	exit 2
fi

# Check for FC transport
numTransports=$(ls -1 /sys/class/fc_transport/ | wc -l)
if [ "$numTransports" = "0" ]
then
	log_err "After doing a scsi bus scan, there are no entries in /sys/class/fc_transport. Check zoning."
	exit 3
fi

log_msg "Number of FC HBAs=$numHosts,  Number of HBAs online=$numOnline,  Number of FC transports=$numTransports"

# Check for file existence to indicate whether a sancheck cleanup run should be done prior to scanning
if [ -f "$vmcdir/stale.sancheck" -a "$argids" = "N" ]
then
	bv=$(cat $vmcdir/stale.sancheck)
	log_msg "  Inventory pre SAN check & fix (staleVols=$bv)"
	sancheck "-fn" "$vmcdir/outinvsancheck.trace" "stale"
	rc=$?
	log_msg "  Inventory pre SAN check complete."
fi

if [ -f $fixlogfile ]
then
	ccrc=$(find $fixlogfile -size +3000)
	if [ "$ccrc" = "$fixlogfile" ]
	then
		log_msg "Cleanup old log file (wrap to backup)."
		mv -f $fixlogfile "${fixlogfile}.1"
	fi
fi

# Cleanup old temp map files older than a week
log_msg "Cleanup old temporary mapping files: "
find $vmcdir/pvmap.* -mtime +7 -exec rm -f {} \; 2> /dev/null
find $vmcdir/devmap.* -mtime +7 -exec rm -f {} \; 2> /dev/null

numvses=$(find /etc/libvirt/qemu -maxdepth 1 -name '*.xml' | wc -l)
#=========================
# Outer loop - fix mode
#=========================
while [ "$fixmode" != "done" ]
do
	log_msg "Begin outer loop. Fixmode=$fixmode.  Pause 2 seconds after scan"
	# First, a short delay
	sleep 2
	if [ "$argids" = "N" ]
	then
		# Specific volumes not passed in, so process all current volumes
		get_vol_inventory
	fi

	let numtimes=0
	firsttime="true"
	# Timeout loop - waiting for devices to appear
	while [ $numtimes -lt $innertries ]
	do
		let numtimes+=1
		log_msg "Begin volume wait loop number $numtimes..."
		if [ "$firsttime" = "false" ]
		then
			log_msg "Pause 10..."
			sleep 10
		fi
		firsttime="false"
		
		find_specified_vols

		# End PV loop
		if [ "$numfound" = "$numvols" ]
		then
		    dstamp=`date`
			log_msg " *** $dstamp *** Number found matches the number sought. Setting fixmode=done"
			fixmode="done"
			break
		fi
	done
	# End Timeout loop
	let innertries=2
	if [ "$fixmode" != "done" ]
	then
		if [ "$fixmode" = "id-enabled" ]
		then
			log_msg "run id list mode fixup"
			if [ "$argids" = "Y" ]
			then
				sancheck "-f --ids $*" $fixlogfile
				rc=$?
				fixmode="full"
			else
				sancheck "-fn --ids${listnotfound}" $fixlogfile
				rc=$?
				fixmode="full"
			fi
		else
			if [ "$fixmode" = "full" ]
			then
				log_msg "run full mode fixup"
				sancheck "-f" $fixlogfile
				rc=$?
				fixmode="none"
			else
				log_msg "Still could not find all the volumes! No more attmepts."
				fixmode="done"
			fi
		fi
	fi
done
# End outer fix loop

if [ "$numfound" != "$numvols" ]
then
	# Error case.
	log_err "The host script was unable to locate one or more FC volumes after searching for the prescribed time and doing device linkage cleanup."
	# check for timeouts
	timeouts=$(echo -e $error_codes | grep -c "137")
	if [ "$timeouts" != "0" ]
	then
		log_err "There where $timeouts timeouts while running commands. Check the health of the FC HBA card, that it is seated properly, and the firmware levels are current."
	fi
	log_msg "All volume(s) NOT found!"
	if [ "$argids" = "N" ]
	then
		# return what we found anyway
		cat $pvmapfile
		cat $pvmapfile >> $alogfile
		sort -b -u --key=1,2 $pvmapfile $vmcdir/pvmap > $vmcdir/pvmap.1
		cp -f $vmcdir/pvmap.1 $vmcdir/pvmap
		rm -f $pvmapfile
		dstamp=$(date)
		log_msg "$dstamp"
		log_msg ""
		exit 0
	fi
	# debug output goes to log file because there is an issue with the amount of data that
	# the subagent can return for the CR case.
	if [ -f $pvmapfile ]
	then
		echo "Dumping $pvmapfile ..." >> $alogfile
		cat $pvmapfile >> $alogfile
		rm -f $pvmapfile
	fi
	log_msg "-------------------------------------------------------"
	log_msg "Dumping sancheck table..."
	sancheck "--ids $*" $alogfile
	log_msg "-------------------------------------------------------"
	log_msg "---------------------FAILED----------------------------"
	dstamp=$(date)
	log_msg "$dstamp"
	exit 4
else
	# Success case
	if [ "$numvols" = "0" ]
	then
		log_print "No LUNs attached to find."
	else
		log_print "All volume(s) found!"
		log_msg "wwid  stable-path  devpath  scsi-addr  lun-in  host-lun  blocks  mpath  access  host-wwpn  vendor  model  external-address  block-size  sizeGiB  health  device-date  using-vs"
	fi
	cat $pvmapfile
	cat $pvmapfile >> $alogfile
	sort -b -u --key=1,2 $pvmapfile $vmcdir/pvmap > $vmcdir/pvmap.1
	cp -f $vmcdir/pvmap.1 $vmcdir/pvmap
	rm -f $pvmapfile
	dstamp=$(date)
	log_msg "$dstamp"
	log_msg ""
	exit 0
fi
# unreachable
